DOCKER_IMAGE_PREFIX = "kestrel-analytics-"

VAR_FILE_SUFFIX = ".parquet.gz"

# folder for input data (Kestrel variables)
# input/argument variables will be 0.parquet.gz, 1.parquet.gz, 2.parquet.gz, etc. in the folder
INPUT_FOLDER = "input"

# folder for output data (Kestrel variables)
# output/returned variables will be 0.parquet.gz, 1.parquet.gz, 2.parquet.gz, etc. in the folder
OUTPUT_FOLDER = "output"

# folder for additional returned display information
# should be one and only one display file (arbitrary name) in the folder with suffix .html or .pickle (Python object)
DISP_FOLDER = "display"



import json
import re
import uuid
import pathlib
import requests

from stix2matcher.matcher import match

from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromFile
from kestrel.exceptions import DataSourceManagerInternalError, DataSourceConnectionError


def _make_query_dir(uri):
    path = pathlib.Path(str(uuid.uuid5(uuid.NAMESPACE_URL, str(uri))))
    path.mkdir(parents=True, exist_ok=True)
    return path


def fixup_pattern(pattern):
    # The matcher doesn't accept TimestampLiterals in START/STOP
    # See https://github.com/oasis-open/cti-pattern-validator/issues/52
    return re.sub(r"(START|STOP)\s+t'", r"\1 '", pattern)


class StixBundleInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        return ["file", "http", "https"]

    @staticmethod
    def list_data_sources():
        """Empty return for now.

        May not be meaningful to implement this for this interface. This method
        should enumerate hosts but the set is unbounded.

        """
        return []

    @staticmethod
    def query(uri, pattern, session_id=None):
        scheme, _, data_paths = uri.rpartition("://")
        data_paths = data_paths.split(",")
        pattern = fixup_pattern(pattern)

        ingestdir = _make_query_dir(uri)
        bundles = []
        for i, data_path in enumerate(data_paths):
            data_path_striped = "".join(filter(str.isalnum, data_path))
            ingestfile = ingestdir / f"{i}_{data_path_striped}.json"

            # TODO: keep files in LRU cache?
            if scheme == "file":
                try:
                    with open(data_path, "r") as f:
                        bundle_in = json.load(f)
                except Exception:
                    raise DataSourceConnectionError(uri)
            elif scheme == "http" or scheme == "https":
                try:
                    bundle_in = requests.get(f"{scheme}://{data_path}").json()
                except requests.exceptions.ConnectionError:
                    raise DataSourceConnectionError(uri)
            else:
                raise DataSourceManagerInternalError(
                    f"interface {__package__} should not process scheme {scheme}"
                )

            bundle_out = {}
            for prop, val in bundle_in.items():
                if prop == "objects":
                    bundle_out[prop] = []
                    for obj in val:
                        if obj["type"] != "observed-data" or match(
                            pattern, [obj], False
                        ):
                            bundle_out[prop].append(obj)
                else:
                    bundle_out[prop] = val

            with ingestfile.open("w") as f:
                json.dump(bundle_out, f)
            bundles.append(str(ingestfile.resolve()))

        return ReturnFromFile(ingestdir.name, bundles)



import logging
import re

from kestrel.exceptions import InvalidAttribute, VariableNotExist, UnsupportedRelation
from kestrel.codegen.relations import stix_2_0_ref_mapping, generic_relations

_logger = logging.getLogger(__name__)


def check_elements_not_empty(stmt):
    for k, v in stmt.items():
        if isinstance(v, str) and not v:
            raise KestrelInternalError(f'incomplete parser; empty value for "{k}"')


def get_entity_table(var_name, symtable):
    if var_name not in symtable:
        raise VariableNotExist(var_name)
    return symtable[var_name].entity_table


def get_entity_type(var_name, symtable):
    if var_name not in symtable:
        raise VariableNotExist(var_name)
    return symtable[var_name].type


def get_entity_len(var_name, symtable):
    if var_name not in symtable:
        raise VariableNotExist(var_name)
    return len(symtable[var_name])


def recognize_var_source(stmt, symtable):
    # parser doesn't understand whether a data source is a Kestrel var
    # this function differente a Kestrel variable source from a data source
    if "datasource" in stmt:
        source = stmt["datasource"]
        if source in symtable:
            stmt["variablesource"] = source
            del stmt["datasource"]


def complete_data_source(stmt, ds):
    if stmt["command"] == "get":
        if "variablesource" not in stmt and "datasource" not in stmt:
            if ds:
                stmt["datasource"] = ds


def normalize_attrs(stmt, v):
    props = []
    for attr in re.split(r",\s*", stmt["attrs"]):
        entity_type, _, prop = attr.rpartition(":")
        if entity_type and entity_type != v.type:
            raise InvalidAttribute(attr)
        props.append(prop)
    return ",".join(props)


def check_semantics_on_find(stmt, input_type):

    if stmt["command"] != "find":
        return

    # relation should be in lowercase after parsing by kestrel.syntax.parser
    relation = stmt["relation"]
    return_type = stmt["type"]

    (entity_x, entity_y) = (
        (input_type, return_type) if stmt["reversed"] else (return_type, input_type)
    )

    if (
        entity_x,
        relation,
        entity_y,
    ) not in stix_2_0_ref_mapping and relation not in generic_relations:
        raise UnsupportedRelation(entity_x, relation, entity_y)


def check_var_exists(var_name, symtable):
    if var_name not in symtable:
        raise VariableNotExist(var_name)



from lark import Lark
from pkgutil import get_data
from itertools import chain

from kestrel.codegen.relations import all_relations


def get_keywords():
    grammar = get_data(__name__, "kestrel.lark").decode("utf-8")
    parser = Lark(grammar, parser="lalr")
    alphabet_patterns = filter(lambda x: x.pattern.value.isalnum(), parser.terminals)
    keywords = [x.pattern.value for x in alphabet_patterns] + all_relations
    keywords_lower = map(lambda x: x.lower(), keywords)
    keywords_upper = map(lambda x: x.upper(), keywords)
    keywords_comprehensive = list(chain(keywords_lower, keywords_upper))
    return keywords_comprehensive



"""Relationship mapping from Kestrel relation to STIX references.

STIX reference names may not be the original STIX reference name. The names
used here are pre-processed by :func:`firepit.raft.invert`. Check the function
for more details.

"""

import dateutil.parser
import datetime
from collections import defaultdict
import logging

from firepit.query import Query, Projection, Table, Unique

_logger = logging.getLogger(__name__)

stix_2_0_ref_mapping = {
    # (EntityX, Relate, EntityY): ([STIX_Ref_i], is_Ref_in_EntityX)
    # All STIX 2.0 refs enumerated
    # file
    ("file", "contained", "artifact"): (["content_ref"], True),
    ("directory", "contained", "directory"): (["x_contains_of_ref"], False),
    ("directory", "contained", "file"): (
        ["x_contains_of_ref", "parent_directory_ref"],
        False,
    ),
    ("archive-ext", "contained", "file"): (["x_contains_of_ref"], False),
    # email
    ("user-account", "owned", "email-addr"): (["belongs_to_ref"], False),
    ("email-addr", "created", "email-message"): (["from_ref", "sender_ref"], False),
    ("email-addr", "accepted", "email-message"): (
        ["x_to_of_ref", "x_cc_of_ref", "x_bcc_of_ref"],
        True,
    ),
    ("email-message", None, "artifact"): (["raw_email_ref", "body_raw_ref"], True),
    ("email-message", None, "file"): (["body_raw_ref"], True),
    # ip address
    ("autonomous-system", "owned", "ipv4-addr"): (["x_belongs_to_of_ref"], True),
    ("autonomous-system", "owned", "ipv6-addr"): (["x_belongs_to_of_ref"], True),
    # network-traffic
    ("ipv4-addr", "created", "network-traffic"): (["src_ref"], False),
    ("ipv6-addr", "created", "network-traffic"): (["src_ref"], False),
    ("mac-addr", "created", "network-traffic"): (["src_ref"], False),
    ("domain-name", "created", "network-traffic"): (["src_ref"], False),
    ("artifact", "created", "network-traffic"): (["src_payload_ref"], False),
    ("mac-addr", None, "ipv4-addr"): (["x_resolves_to_of_ref"], True),
    ("mac-addr", None, "ipv6-addr"): (["x_resolves_to_of_ref"], True),
    ("http-request-ext", None, "artifact"): (["message_body_data_ref"], True),
    ("ipv4-addr", "accepted", "network-traffic"): (["dst_ref"], False),
    ("ipv6-addr", "accepted", "network-traffic"): (["dst_ref"], False),
    ("mac-addr", "accepted", "network-traffic"): (["dst_ref"], False),
    ("domain-name", "accepted", "network-traffic"): (["dst_ref"], False),
    ("artifact", "accepted", "network-traffic"): (["dst_payload_ref"], False),
    ("network-traffic", "contained", "network-traffic"): (
        ["encapsulated_by_ref"],
        False,
    ),
    # process
    ("process", "created", "network-traffic"): (["x_opened_connection_of_ref"], False),
    ("user-account", "owned", "process"): (["creator_user_ref"], False),
    ("process", "loaded", "file"): (["binary_ref"], True),
    ("process", "created", "process"): (["parent_ref"], False),
    # service
    ("windows-service-ext", "loaded", "file"): (["x_service_dll_of_ref"], False),
    ("windows-service-ext", "loaded", "user-account"): (["creator_user_ref"], True),
}

# the first available attribute will be used to uniquely identify the entity
stix_2_0_identical_mapping = {
    # entity-type: id attributes candidates
    "directory": ("path",),
    "domain-name": ("value",),
    "email-addr": ("value",),
    "file": ("name",),  # optional in STIX standard
    "ipv4-addr": ("value",),
    "ipv6-addr": ("value",),
    "mac-addr": ("value",),
    "mutex": ("name",),
    # `pid` is optional in STIX standard
    # `first_observed` cannot be used since it may be wrong (derived from observation)
    # `command_line` or `name` may not be in data and cannot be used
    "process": ("pid", "name"),
    "software": ("name",),
    "url": ("value",),
    "user-account": ("user_id",),  # optional in STIX standard
    "windows-registry-key": ("key",),  # optional in STIX standard
}

stix_x_ibm_event_mapping = {
    # entity-type to ref in x-oca-event
    "process": "process_ref",
    "domain-name": "domain_ref",
    "file": "file_ref",
    "user-account": "user_ref",
    "windows-registry-key": "registry_ref",
    "network-traffic": "nt_ref",
    "x-oca-asset": "host_ref",
}

# no direction for generic relations
generic_relations = ["linked"]

all_relations = list(
    set([x[1] for x in stix_2_0_ref_mapping.keys() if x[1]] + generic_relations)
)


def get_entity_id_attribute(variable):
    # this function should always return something
    # if no entity id attribute found, fall back to record "id" by default
    # this works for:
    #   - no appriparite identifier attribute found given specific data
    #   - "network-traffic" (not in stix_2_0_identical_mapping)
    id_attr = "id"

    if variable.type in stix_2_0_identical_mapping:
        available_attributes = variable.store.columns(variable.entity_table)
        for attr in stix_2_0_identical_mapping[variable.type]:
            if attr in available_attributes:
                query = Query()
                query.append(Table(variable.entity_table))
                query.append(Projection([attr]))
                query.append(Unique())
                rows = variable.store.run_query(query).fetchall()
                all_values = [row[attr] for row in rows if row[attr]]
                if all_values:
                    id_attr = attr
                    break

    return id_attr


def are_entities_associated_with_x_ibm_event(entity_types):
    flags = [entity_type in stix_x_ibm_event_mapping for entity_type in entity_types]
    return all(flags)


def compile_generic_relation_to_pattern(return_type, input_type, input_var_name):
    comp_exps = []
    for relation, is_reversed in _enumerate_relations_between_entities(
        return_type, input_type
    ):
        comp_exps += _generate_paramstix_comparison_expressions(
            return_type, relation, input_type, is_reversed, input_var_name
        )
    pattern = "[" + " OR ".join(comp_exps) + "]"
    _logger.debug(f"generic relation pattern compiled: {pattern}")
    return pattern


def compile_specific_relation_to_pattern(
    return_type, relation, input_type, is_reversed, input_var_name
):
    comp_exps = _generate_paramstix_comparison_expressions(
        return_type, relation, input_type, is_reversed, input_var_name
    )
    pattern = "[" + " OR ".join(comp_exps) + "]"
    _logger.debug(f"specific relation pattern compiled: {pattern}")
    return pattern


def compile_identical_entity_search_pattern(var_name, var_struct, does_support_id):
    # "id" attribute may not be available for STIX 2.0 via STIX-shifter
    # so `does_support_id` is set to False in default kestrel config file
    attribute = get_entity_id_attribute(var_struct)
    if attribute == "id" and not does_support_id:
        pattern = None
    else:
        pattern = f"[{var_struct.type}:{attribute} = {var_name}.{attribute}]"
    _logger.debug(f"identical entity search pattern compiled: {pattern}")
    return pattern


def compile_x_ibm_event_search_flow_in_pattern(input_type, input_var_name):
    ref = stix_x_ibm_event_mapping[input_type]
    pattern = f"[x-oca-event:{ref}.id = {input_var_name}.id]"
    _logger.debug(f"x-oca-event flow in pattern compiled: {pattern}")
    return pattern


def compile_x_ibm_event_search_flow_out_pattern(return_type, input_event_var_name):
    ref = stix_x_ibm_event_mapping[return_type]
    pattern = f"[{return_type}:id = {input_event_var_name}.{ref}.id]"
    _logger.debug(f"x-oca-event flow out pattern compiled: {pattern}")
    return pattern


def _enumerate_relations_between_entities(return_type, input_type):
    # return: [(relation, is_reversed)]
    relations = []
    for (x, r, y) in stix_2_0_ref_mapping.keys():
        if x == return_type and y == input_type:
            relations.append((r, False))
        if y == return_type and x == input_type:
            relations.append((r, True))
    _logger.debug(
        f'enumerated relations between "{return_type}" and "{input_type}": {relations}'
    )
    return relations


def _generate_paramstix_comparison_expressions(
    return_type, relation, input_type, is_reversed, input_var_name
):
    (entity_x, entity_y) = (
        (input_type, return_type) if is_reversed else (return_type, input_type)
    )

    stix_refs, is_ref_in_entityx = stix_2_0_ref_mapping[(entity_x, relation, entity_y)]

    is_ref_in_return_entity = is_reversed ^ is_ref_in_entityx

    comp_exps = []
    for stix_ref in stix_refs:
        if is_ref_in_return_entity:
            comp_exps.append(f"{return_type}:{stix_ref}.id = {input_var_name}.id")
        else:
            comp_exps.append(f"{return_type}:id = {input_var_name}.{stix_ref}.id")

    return comp_exps


def fine_grained_relational_process_filtering(
    local_var, prefetch_entity_table, store, config
):

    _logger.debug(
        f"start fine-grained relational process filtering for prefetched table: {prefetch_entity_table}"
    )

    query_ref = Query()
    query_ref.append(Table(local_var.entity_table))
    query_ref.append(Projection(["pid", "name", "first_observed", "last_observed"]))
    ref_rows = local_var.store.run_query(query_ref).fetchall()

    entities = defaultdict(list)

    for row in ref_rows:
        if row["pid"]:
            process_name = row["name"]
            process_start_time = dateutil.parser.isoparse(row["first_observed"])
            process_end_time = dateutil.parser.isoparse(row["last_observed"])
            entities[row["pid"]].append(
                (process_name, process_start_time, process_end_time)
            )

    query_fil = Query()
    query_fil.append(Table(prefetch_entity_table))
    query_fil.append(
        Projection(["id", "pid", "name", "first_observed", "last_observed"])
    )
    fil_rows = store.run_query(query_fil).fetchall()

    filtered_ids = []

    pnc_start_offset = datetime.timedelta(
        seconds=config["process_name_change_timerange_start_offset"]
    )
    pnc_stop_offset = datetime.timedelta(
        seconds=config["process_name_change_timerange_stop_offset"]
    )
    pls_start_offset = datetime.timedelta(
        seconds=config["process_lifespan_start_offset"]
    )
    pls_stop_offset = datetime.timedelta(seconds=config["process_lifespan_stop_offset"])

    for row in fil_rows:
        if row["pid"]:
            fil_start_time = dateutil.parser.isoparse(row["first_observed"])
            fil_end_time = dateutil.parser.isoparse(row["last_observed"])
            fil_process_name = row["name"]
            for ref_process_name, ref_start_time, ref_end_time in entities[row["pid"]]:
                if (
                    (
                        fil_process_name
                        and fil_process_name == ref_process_name
                        and fil_start_time > ref_start_time + pls_start_offset
                        and fil_start_time < ref_end_time + pls_stop_offset
                    )
                    or (
                        fil_start_time > ref_start_time + pnc_start_offset
                        and fil_start_time < ref_end_time + pnc_stop_offset
                    )
                    or (
                        fil_end_time > ref_start_time + pnc_start_offset
                        and fil_end_time < ref_end_time + pnc_stop_offset
                    )
                ):
                    filtered_ids.append(row["id"])
                    break

    _logger.debug(
        f"found {len(filtered_ids)} out of {len(fil_rows)} to be true relational process records."
    )

    return filtered_ids



"""STIX Shifter data source package provides access to data sources via
stix-shifter.

Before use, need to install the target stix-shifter connector packages such as
``stix-shifter-modules-carbonblack``.

The STIX Shifter interface can reach multiple data sources. The user needs to
setup one *profile* per data source. The profile name will be used in the
``FROM`` clause of the Kestrel ``GET`` command, e.g., ``newvar = GET entity-type
FROM stixshifter://profilename WHERE ...``. Kestrel runtime will load the profile
for the used profile from environment variables:

- ``STIXSHIFTER_PROFILENAME_CONNECTOR``: the STIX Shifter connector name, e.g., ``elastic_ecs``.
- ``STIXSHIFTER_PROFILENAME_CONNECTION``: the STIX Shifter `connection <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#connection>`_ object in JSON string.
- ``STIXSHIFTER_PROFILENAME_CONFIG``: the STIX Shifter `configuration <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#configuration>`_ object in JSON string.

Properties of profile name:

- Not case sensitive, e.g., ``profileX`` in the Kestrel command will match
  ``STIXSHIFTER_PROFILEX_...`` in environment variables.
- Cannot contain ``_``.

"""

import os
import json
import time

from stix_shifter.stix_translation import stix_translation
from stix_shifter.stix_transmission import stix_transmission

from kestrel.utils import mkdtemp
from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromFile
from kestrel.exceptions import (
    InvalidDataSource,
    DataSourceError,
    DataSourceManagerInternalError,
)

from kestrel_datasource_stixshifter.config import ENV_VAR_PREFIX, RETRIEVAL_BATCH_SIZE


class StixShifterInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        """STIX Shifter data source interface only supports ``stixshifter://`` scheme."""
        return ["stixshifter"]

    @staticmethod
    def list_data_sources():
        """Get configured data sources from environment variable profiles."""
        data_sources = []

        env_vars = os.environ.keys()
        stixshifter_vars = filter(lambda x: x.startswith(ENV_VAR_PREFIX), env_vars)
        for evar in stixshifter_vars:
            profile = evar.split("_")[1].lower()
            if profile not in data_sources:
                data_sources.append(profile)

        data_sources.sort()

        return data_sources

    @staticmethod
    def query(uri, pattern, session_id=None):
        """Query a stixshifter data source."""
        scheme, _, profile = uri.rpartition("://")
        profiles = profile.split(",")

        if scheme != "stixshifter":
            raise DataSourceManagerInternalError(
                f"interface {__package__} should not process scheme {scheme}"
            )

        ingestdir = mkdtemp()
        query_id = ingestdir.name
        bundles = []
        for i, profile in enumerate(profiles):
            (
                connector_name,
                connection_dict,
                configuration_dict,
            ) = StixShifterInterface._get_stixshifter_config(profile)

            data_path_striped = "".join(filter(str.isalnum, profile))
            ingestfile = ingestdir / f"{i}_{data_path_striped}.json"

            query_metadata = json.dumps(
                {"id": "identity--" + query_id, "name": connector_name}
            )

            translation = stix_translation.StixTranslation()
            transmission = stix_transmission.StixTransmission(
                connector_name, connection_dict, configuration_dict
            )

            dsl = translation.translate(
                connector_name, "query", query_metadata, pattern, {}
            )

            if "error" in dsl:
                raise DataSourceError(
                    f"STIX-shifter translation failed with message: {dsl['error']}"
                )

            # query results should be put together; when translated to STIX, the relation between them will remain
            connector_results = []
            for query in dsl["queries"]:
                search_meta_result = transmission.query(query)
                if search_meta_result["success"]:
                    search_id = search_meta_result["search_id"]
                    if transmission.is_async():
                        time.sleep(1)
                        status = transmission.status(search_id)
                        if status["success"]:
                            while (
                                status["progress"] < 100
                                and status["status"] == "RUNNING"
                            ):
                                status = transmission.status(search_id)
                        else:
                            stix_shifter_error_msg = (
                                status["error"]
                                if "error" in status
                                else "details not avaliable"
                            )
                            raise DataSourceError(
                                f"STIX-shifter transmission.status() failed with message: {stix_shifter_error_msg}"
                            )

                    result_retrieval_offset = 0
                    has_remaining_results = True
                    while has_remaining_results:
                        result_batch = transmission.results(
                            search_id, result_retrieval_offset, RETRIEVAL_BATCH_SIZE
                        )
                        if result_batch["success"]:
                            new_entries = result_batch["data"]
                            if new_entries:
                                connector_results += new_entries
                                result_retrieval_offset += RETRIEVAL_BATCH_SIZE
                                if len(new_entries) < RETRIEVAL_BATCH_SIZE:
                                    has_remaining_results = False
                            else:
                                has_remaining_results = False
                        else:
                            stix_shifter_error_msg = (
                                result_batch["error"]
                                if "error" in result_batch
                                else "details not avaliable"
                            )
                            raise DataSourceError(
                                f"STIX-shifter transmission.results() failed with message: {stix_shifter_error_msg}"
                            )

                else:
                    stix_shifter_error_msg = (
                        search_meta_result["error"]
                        if "error" in search_meta_result
                        else "details not avaliable"
                    )
                    raise DataSourceError(
                        f"STIX-shifter transmission.query() failed with message: {stix_shifter_error_msg}"
                    )

            stixbundle = translation.translate(
                connector_name,
                "results",
                query_metadata,
                json.dumps(connector_results),
                {},
            )

            with ingestfile.open("w") as ingest:
                json.dump(stixbundle, ingest, indent=4)
            bundles.append(str(ingestfile.resolve()))

        return ReturnFromFile(query_id, bundles)

    @staticmethod
    def _get_stixshifter_config(profile_name):
        profile_name = profile_name.upper()

        env_conr_name = f"{ENV_VAR_PREFIX}{profile_name}_CONNECTOR"
        connector_name = os.getenv(env_conr_name)
        if not connector_name:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"no {env_conr_name} environment variable found",
            )
        connector_name = connector_name.lower()

        env_conn_name = f"{ENV_VAR_PREFIX}{profile_name}_CONNECTION"
        connection = os.getenv(env_conn_name)
        if not connection:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"no {env_conn_name} environment variable found",
            )

        env_conf_name = f"{ENV_VAR_PREFIX}{profile_name}_CONFIG"
        configuration = os.getenv(env_conf_name)
        if not configuration:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"no {env_conf_name} environment variable found",
            )

        try:
            connection = json.loads(connection)
        except json.decoder.JSONDecodeError:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"invalid JSON in {env_conn_name} environment variable",
            )

        if "host" not in connection:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f'invalid {env_conn_name} environment variable: no "host" field',
            )

        if "port" not in connection and connector_name != "stix_bundle":
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f'invalid {env_conn_name} environment variable: no "port" field',
            )

        try:
            configuration = json.loads(configuration)
        except json.decoder.JSONDecodeError:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f"invalid JSON in {env_conf_name} environment variable",
            )

        if "auth" not in configuration:
            raise InvalidDataSource(
                profile_name,
                "stixshifter",
                f'invalid {env_conf_name} environment variable: no "auth" field',
            )

        return connector_name, connection, configuration



from abc import ABC, abstractmethod


class AbstractReturnStruct(ABC):
    """The abstract class for creating return objects.

    1. it should have a constructor for the interface to create it. The
       interface should specify the ``query_id`` in the constructor.

    2. it should have a :attr:`load_to_store` method for Kestrel runtime to load data from it.

    """

    @abstractmethod
    def load_to_store(self, store):
        """Load the data (from data source) to store.

        Returns:
            str: the ``query_id``, which is a identifier in the store
            associated with the loaded data entries.

        """
        return ""


class ReturnFromFile(AbstractReturnStruct):
    """The return structure when the data source interface uses files as intermediate storage before loading to store.

    Args:
        query_id (str): typically just a UUID.
        file_paths ([str]): the list of stix bundle file paths.

    """

    def __init__(self, query_id, file_paths):
        self.query_id = query_id
        self.file_paths = file_paths

    def load_to_store(self, store):
        store.cache(self.query_id, self.file_paths)
        return self.query_id


class ReturnFromStore(AbstractReturnStruct):
    """The return structure when the data source interface directly operates on the store.

    Args:
        query_id (str): typically just a UUID.

    """

    def __init__(self, query_id):
        self.query_id = query_id

    def load_to_store(self, store):
        return self.query_id
